@@ -56,6 +56,7 @@ impl OutputId {
 
 impl Backend {
     pub fn init(&mut self, niri: &mut Niri) {
+        let _span = tracy_client::span!("Backend::init");
         match self {
             Backend::Tty(tty) => tty.init(niri),
             Backend::Winit(winit) => winit.init(niri),
@@ -184,6 +185,14 @@ impl Backend {
         }
     }
 
+    pub fn update_ignored_nodes_config(&mut self, niri: &mut Niri) {
+        match self {
+            Backend::Tty(tty) => tty.update_ignored_nodes_config(niri),
+            Backend::Winit(_) => (),
+            Backend::Headless(_) => (),
+        }
+    }
+
     pub fn on_output_config_changed(&mut self, niri: &mut Niri) {
         match self {
             Backend::Tty(tty) => tty.on_output_config_changed(niri),
