@@ -14,13 +14,13 @@ use smithay::wayland::compositor::{
     SurfaceAttributes,
 };
 use smithay::wayland::dmabuf::get_dmabuf;
-use smithay::wayland::shell::xdg::XdgToplevelSurfaceData;
+use smithay::wayland::shell::xdg::ToplevelCachedState;
 use smithay::wayland::shm::{ShmHandler, ShmState};
 use smithay::{delegate_compositor, delegate_shm};
 
 use super::xdg_shell::add_mapped_toplevel_pre_commit_hook;
 use crate::handlers::XDG_ACTIVATION_TOKEN_TIMEOUT;
-use crate::layout::{ActivateWindow, AddWindowTarget};
+use crate::layout::{ActivateWindow, AddWindowTarget, LayoutElement as _};
 use crate::niri::{CastTarget, ClientState, LockState, State};
 use crate::utils::transaction::Transaction;
 use crate::utils::{is_mapped, send_scale_transform};
@@ -56,7 +56,8 @@ impl CompositorHandler for State {
 
     fn commit(&mut self, surface: &WlSurface) {
         let _span = tracy_client::span!("CompositorHandler::commit");
-        trace!(surface = ?surface.id(), "commit");
+        let _span = trace_span!("commit", surface = %surface.id()).entered();
+        trace!("commit");
 
         on_commit_buffer_handler::<Self>(surface);
         self.backend.early_import(surface);
@@ -90,33 +91,59 @@ impl CompositorHandler for State {
 
                     let toplevel = window.toplevel().expect("no X11 support");
 
-                    let (rules, width, height, is_full_width, output, workspace_id) =
-                        if let InitialConfigureState::Configured {
+                    let (
+                        rules,
+                        width,
+                        height,
+                        is_full_width,
+                        output,
+                        workspace_id,
+                        is_pending_maximized,
+                    ) = if let InitialConfigureState::Configured {
+                        rules,
+                        width,
+                        height,
+                        floating_width: _,
+                        floating_height: _,
+                        is_full_width,
+                        output,
+                        workspace_name,
+                        is_pending_maximized,
+                    } = state
+                    {
+                        // Check that the output is still connected.
+                        let output =
+                            output.filter(|o| self.niri.layout.monitor_for_output(o).is_some());
+
+                        // Check that the workspace still exists.
+                        let workspace_id = workspace_name
+                            .as_deref()
+                            .and_then(|n| self.niri.layout.find_workspace_by_name(n))
+                            .map(|(_, ws)| ws.id());
+
+                        (
                             rules,
                             width,
                             height,
-                            floating_width: _,
-                            floating_height: _,
                             is_full_width,
                             output,
-                            workspace_name,
-                        } = state
-                        {
-                            // Check that the output is still connected.
-                            let output =
-                                output.filter(|o| self.niri.layout.monitor_for_output(o).is_some());
-
-                            // Check that the workspace still exists.
-                            let workspace_id = workspace_name
-                                .as_deref()
-                                .and_then(|n| self.niri.layout.find_workspace_by_name(n))
-                                .map(|(_, ws)| ws.id());
-
-                            (rules, width, height, is_full_width, output, workspace_id)
-                        } else {
-                            error!("window map must happen after initial configure");
-                            (ResolvedWindowRules::empty(), None, None, false, None, None)
-                        };
+                            workspace_id,
+                            is_pending_maximized,
+                        )
+                    } else {
+                        // Can happen when a surface unmaps by attaching a null buffer while
+                        // there are in-flight pending configures.
+                        debug!("window mapped without proper initial configure");
+                        (
+                            ResolvedWindowRules::default(),
+                            None,
+                            None,
+                            false,
+                            None,
+                            None,
+                            false,
+                        )
+                    };
 
                     // The GTK about dialog sets min/max size after the initial configure but
                     // before mapping, so we need to compute open_floating at the last possible
@@ -166,7 +193,7 @@ impl CompositorHandler for State {
                         .map(|(mapped, _)| mapped.window.clone());
 
                     // The mapped pre-commit hook deals with dma-bufs on its own.
-                    self.remove_default_dmabuf_pre_commit_hook(toplevel.wl_surface());
+                    self.remove_default_dmabuf_pre_commit_hook(surface);
                     let hook = add_mapped_toplevel_pre_commit_hook(toplevel);
                     let mapped = Mapped::new(window, rules, hook);
                     let window = mapped.window.clone();
@@ -190,8 +217,21 @@ impl CompositorHandler for State {
                         is_floating,
                         activate,
                     );
+                    let output = output.cloned();
 
-                    if let Some(output) = output.cloned() {
+                    // The window state cannot contain Fullscreen and Maximized at once. Therefore,
+                    // if the window ended up fullscreen, then we only know that it is also
+                    // maximized from the is_pending_maximized variable. Tell the layout about it
+                    // here so that unfullscreening the window makes it maximized.
+                    if let Some((mapped, _)) = self.niri.layout.find_window_and_output(surface) {
+                        if mapped.pending_sizing_mode().is_fullscreen() && is_pending_maximized {
+                            self.niri.layout.set_maximized(&window, true);
+                        }
+                    } else {
+                        error!("layout is missing the window that we just added");
+                    }
+
+                    if let Some(output) = output {
                         self.niri.layout.start_open_animation_for_window(&window);
 
                         let new_focus = self.niri.layout.focus().map(|m| &m.window);
@@ -207,6 +247,7 @@ impl CompositorHandler for State {
                 }
 
                 // The toplevel remains unmapped.
+                trace!("toplevel remains unmapped");
                 let unmapped = entry.get();
                 if unmapped.needs_initial_configure() {
                     let toplevel = unmapped.window.toplevel().expect("no x11 support").clone();
@@ -242,12 +283,15 @@ impl CompositorHandler for State {
                     // The toplevel got unmapped.
                     //
                     // Test client: wleird-unmap.
+                    trace!("toplevel got unmapped");
+
                     let active_window = self.niri.layout.focus().map(|m| &m.window);
                     let was_active = active_window == Some(&window);
 
                     self.niri
                         .stop_casts_for_target(CastTarget::Window { id: id.get() });
 
+                    self.niri.window_mru_ui.remove_window(id);
                     self.niri.layout.remove_window(&window, transaction.clone());
                     self.add_default_dmabuf_pre_commit_hook(surface);
 
@@ -268,6 +312,7 @@ impl CompositorHandler for State {
 
                     if let Some(output) = output {
                         self.niri.queue_redraw(&output);
+                        self.niri.queue_redraw_mru_output();
                     }
                     return;
                 }
@@ -280,19 +325,21 @@ impl CompositorHandler for State {
                         .buffer_delta
                         .take();
 
-                    let role = states
-                        .data_map
-                        .get::<XdgToplevelSurfaceData>()
-                        .unwrap()
-                        .lock()
-                        .unwrap();
-                    (role.configure_serial, buffer_delta)
+                    let serial = states
+                        .cached_state
+                        .get::<ToplevelCachedState>()
+                        .current()
+                        .last_acked
+                        .as_ref()
+                        .map(|c| c.serial);
+                    (serial, buffer_delta)
                 });
                 if serial.is_none() {
                     error!("commit on a mapped surface without a configured serial");
                 }
 
                 // The toplevel remains mapped.
+                self.niri.window_mru_ui.update_window(&self.niri.layout, id);
                 self.niri.layout.update_window(&window, serial);
 
                 // Move the toplevel according to the attach offset.
@@ -313,6 +360,7 @@ impl CompositorHandler for State {
 
                 if let Some(output) = output {
                     self.niri.queue_redraw(&output);
+                    self.niri.queue_redraw_mru_output();
                 }
                 return;
             }
@@ -326,9 +374,13 @@ impl CompositorHandler for State {
             let window = mapped.window.clone();
             let output = output.cloned();
             window.on_commit();
+            self.niri
+                .window_mru_ui
+                .update_window(&self.niri.layout, mapped.id());
             self.niri.layout.update_window(&window, None);
             if let Some(output) = output {
                 self.niri.queue_redraw(&output);
+                self.niri.queue_redraw_mru_output();
             }
             return;
         }
