@@ -2,7 +2,7 @@ use std::cell::RefCell;
 use std::collections::HashSet;
 use std::ffi::OsStr;
 use std::os::unix::net::{UnixListener, UnixStream};
-use std::path::PathBuf;
+use std::path::{Path, PathBuf};
 use std::rc::Rc;
 use std::sync::{Arc, Mutex};
 use std::{env, io, process};
@@ -17,8 +17,8 @@ use futures_util::{select_biased, AsyncBufReadExt, AsyncWrite, AsyncWriteExt, Fu
 use niri_config::OutputName;
 use niri_ipc::state::{EventStreamState, EventStreamStatePart as _};
 use niri_ipc::{
-    Event, KeyboardLayouts, OutputConfigChanged, Overview, Reply, Request, Response, WindowLayout,
-    Workspace,
+    Action, Event, KeyboardLayouts, OutputConfigChanged, Overview, Reply, Request, Response,
+    Timestamp, WindowLayout, Workspace,
 };
 use smithay::desktop::layer_map_for_output;
 use smithay::input::pointer::{
@@ -379,6 +379,8 @@ async fn process(ctx: &ClientCtx, request: Request) -> Reply {
             Response::PickedColor(color)
         }
         Request::Action(action) => {
+            validate_action(&action)?;
+
             let (tx, rx) = async_channel::bounded(1);
 
             let action = niri_config::Action::from(action);
@@ -397,6 +399,8 @@ async fn process(ctx: &ClientCtx, request: Request) -> Reply {
             Response::Handled
         }
         Request::Output { output, action } => {
+            action.validate()?;
+
             let ipc_outputs = ctx.ipc_outputs.lock().unwrap();
             let found = ipc_outputs
                 .values()
@@ -451,6 +455,23 @@ async fn process(ctx: &ClientCtx, request: Request) -> Reply {
     Ok(response)
 }
 
+fn validate_action(action: &Action) -> Result<(), String> {
+    if let Action::Screenshot { path, .. }
+    | Action::ScreenshotScreen { path, .. }
+    | Action::ScreenshotWindow { path, .. } = action
+    {
+        if let Some(path) = path {
+            // Relative paths are resolved against the niri compositor's working directory, which
+            // is almost certainly not what you want.
+            if !Path::new(path).is_absolute() {
+                return Err(format!("path must be absolute: {path}"));
+            }
+        }
+    }
+
+    Ok(())
+}
+
 async fn handle_event_stream_client(client: EventStreamClient) -> anyhow::Result<()> {
     let EventStreamClient {
         events,
@@ -493,6 +514,7 @@ fn make_ipc_window(
         is_floating: mapped.is_floating(),
         is_urgent: mapped.is_urgent(),
         layout,
+        focus_timestamp: mapped.get_focus_timestamp().map(Timestamp::from),
     })
 }
 
@@ -704,6 +726,14 @@ impl State {
                 events.push(Event::WindowFocusChanged { id: Some(id) });
             }
 
+            let focus_timestamp = mapped.get_focus_timestamp().map(Timestamp::from);
+            if focus_timestamp != ipc_win.focus_timestamp {
+                events.push(Event::WindowFocusTimestampChanged {
+                    id,
+                    focus_timestamp,
+                });
+            }
+
             let urgent = mapped.is_urgent();
             if urgent != ipc_win.is_urgent {
                 events.push(Event::WindowUrgencyChanged { id, urgent })
@@ -773,4 +803,15 @@ impl State {
         state.apply(event.clone());
         server.send_event(event);
     }
+
+    pub fn ipc_screenshot_taken(&mut self, path: Option<String>) {
+        let Some(server) = &self.niri.ipc_server else {
+            return;
+        };
+        let mut state = server.event_stream_state.borrow_mut();
+
+        let event = Event::ScreenshotCaptured { path };
+        state.apply(event.clone());
+        server.send_event(event);
+    }
 }
