@@ -1,6 +1,7 @@
 use core::f64;
 use std::rc::Rc;
 
+use niri_config::utils::MergeWith as _;
 use niri_config::{Color, CornerRadius, GradientInterpolation};
 use niri_ipc::WindowLayout;
 use smithay::backend::renderer::element::{Element, Kind};
@@ -15,6 +16,7 @@ use super::{
     SizeFrac, RESIZE_ANIMATION_THRESHOLD,
 };
 use crate::animation::{Animation, Clock};
+use crate::layout::SizingMode;
 use crate::niri_render_elements;
 use crate::render_helpers::border::BorderRenderElement;
 use crate::render_helpers::clipped_surface::{ClippedSurfaceRenderElement, RoundedCornerDamage};
@@ -27,7 +29,9 @@ use crate::render_helpers::snapshot::RenderSnapshot;
 use crate::render_helpers::solid_color::{SolidColorBuffer, SolidColorRenderElement};
 use crate::render_helpers::RenderTarget;
 use crate::utils::transaction::Transaction;
-use crate::utils::{baba_is_float_offset, round_logical_in_physical};
+use crate::utils::{
+    baba_is_float_offset, round_logical_in_physical, round_logical_in_physical_max1,
+};
 
 /// Toplevel window with decorations.
 #[derive(Debug)]
@@ -44,17 +48,17 @@ pub struct Tile<W: LayoutElement> {
     /// The shadow around the window.
     shadow: Shadow,
 
-    /// Whether this tile is fullscreen.
+    /// This tile's current sizing mode.
     ///
-    /// This will update only when the `window` actually goes fullscreen, rather than right away,
-    /// to avoid black backdrop flicker before the window has had a chance to resize.
-    is_fullscreen: bool,
+    /// This will update only when the `window` actually goes maximized or fullscreen, rather than
+    /// right away, to avoid black backdrop flicker before the window has had a chance to resize.
+    sizing_mode: SizingMode,
 
     /// The black backdrop for fullscreen windows.
     fullscreen_backdrop: SolidColorBuffer,
 
     /// Whether the tile should float upon unfullscreening.
-    pub(super) unfullscreen_to_floating: bool,
+    pub(super) restore_to_floating: bool,
 
     /// The size that the window should assume when going floating.
     ///
@@ -138,6 +142,15 @@ struct ResizeAnimation {
     size_from: Size<f64, Logical>,
     snapshot: LayoutElementRenderSnapshot,
     offscreen: OffscreenBuffer,
+    tile_size_from: Size<f64, Logical>,
+    // If the resize involved the fullscreen state at some point, this is the progress toward the
+    // fullscreen state. Used for things like fullscreen backdrop alpha.
+    //
+    // Note that this can be set even if this specific resize is between two non-fullscreen states,
+    // for example when issuing a new resize during an unfullscreen resize.
+    fullscreen_progress: Option<Animation>,
+    // Similar to above but for fullscreen-or-maximized.
+    expanded_progress: Option<Animation>,
 }
 
 #[derive(Debug)]
@@ -167,19 +180,19 @@ impl<W: LayoutElement> Tile<W> {
         options: Rc<Options>,
     ) -> Self {
         let rules = window.rules();
-        let border_config = rules.border.resolve_against(options.border);
-        let focus_ring_config = rules.focus_ring.resolve_against(options.focus_ring.into());
-        let shadow_config = rules.shadow.resolve_against(options.shadow);
-        let is_fullscreen = window.is_fullscreen();
+        let border_config = options.layout.border.merged_with(&rules.border);
+        let focus_ring_config = options.layout.focus_ring.merged_with(&rules.focus_ring);
+        let shadow_config = options.layout.shadow.merged_with(&rules.shadow);
+        let sizing_mode = window.sizing_mode();
 
         Self {
             window,
             border: FocusRing::new(border_config.into()),
-            focus_ring: FocusRing::new(focus_ring_config.into()),
+            focus_ring: FocusRing::new(focus_ring_config),
             shadow: Shadow::new(shadow_config),
-            is_fullscreen,
-            fullscreen_backdrop: SolidColorBuffer::new(view_size, [0., 0., 0., 1.]),
-            unfullscreen_to_floating: false,
+            sizing_mode,
+            fullscreen_backdrop: SolidColorBuffer::new((0., 0.), [0., 0., 0., 1.]),
+            restore_to_floating: false,
             floating_window_size: None,
             floating_pos: None,
             floating_preset_width_idx: None,
@@ -206,10 +219,10 @@ impl<W: LayoutElement> Tile<W> {
         options: Rc<Options>,
     ) {
         // If preset widths or heights changed, clear our stored preset index.
-        if self.options.preset_column_widths != options.preset_column_widths {
+        if self.options.layout.preset_column_widths != options.layout.preset_column_widths {
             self.floating_preset_width_idx = None;
         }
-        if self.options.preset_window_heights != options.preset_window_heights {
+        if self.options.layout.preset_window_heights != options.layout.preset_window_heights {
             self.floating_preset_height_idx = None;
         }
 
@@ -217,20 +230,24 @@ impl<W: LayoutElement> Tile<W> {
         self.scale = scale;
         self.options = options;
 
+        let round_max1 = |logical| round_logical_in_physical_max1(self.scale, logical);
+
         let rules = self.window.rules();
 
-        let border_config = rules.border.resolve_against(self.options.border);
+        let mut border_config = self.options.layout.border.merged_with(&rules.border);
+        border_config.width = round_max1(border_config.width);
         self.border.update_config(border_config.into());
 
-        let focus_ring_config = rules
+        let mut focus_ring_config = self
+            .options
+            .layout
             .focus_ring
-            .resolve_against(self.options.focus_ring.into());
-        self.focus_ring.update_config(focus_ring_config.into());
+            .merged_with(&rules.focus_ring);
+        focus_ring_config.width = round_max1(focus_ring_config.width);
+        self.focus_ring.update_config(focus_ring_config);
 
-        let shadow_config = rules.shadow.resolve_against(self.options.shadow);
+        let shadow_config = self.options.layout.shadow.merged_with(&rules.shadow);
         self.shadow.update_config(shadow_config);
-
-        self.fullscreen_backdrop.resize(view_size);
     }
 
     pub fn update_shaders(&mut self) {
@@ -240,28 +257,94 @@ impl<W: LayoutElement> Tile<W> {
     }
 
     pub fn update_window(&mut self) {
-        self.is_fullscreen = self.window.is_fullscreen();
+        let prev_sizing_mode = self.sizing_mode;
+        self.sizing_mode = self.window.sizing_mode();
 
         if let Some(animate_from) = self.window.take_animation_snapshot() {
-            let (size_from, offscreen) = if let Some(resize) = self.resize_animation.take() {
+            let params = if let Some(resize) = self.resize_animation.take() {
                 // Compute like in animated_window_size(), but using the snapshot geometry (since
                 // the current one is already overwritten).
                 let mut size = animate_from.size;
 
                 let val = resize.anim.value();
                 let size_from = resize.size_from;
+                let tile_size_from = resize.tile_size_from;
 
                 size.w = size_from.w + (size.w - size_from.w) * val;
                 size.h = size_from.h + (size.h - size_from.h) * val;
 
+                let mut tile_size = animate_from.size;
+                if prev_sizing_mode.is_fullscreen() {
+                    tile_size.w = f64::max(tile_size.w, self.view_size.w);
+                    tile_size.h = f64::max(tile_size.h, self.view_size.h);
+                } else if prev_sizing_mode.is_normal() && !self.border.is_off() {
+                    let width = self.border.width();
+                    tile_size.w += width * 2.;
+                    tile_size.h += width * 2.;
+                }
+
+                tile_size.w = tile_size_from.w + (tile_size.w - tile_size_from.w) * val;
+                tile_size.h = tile_size_from.h + (tile_size.h - tile_size_from.h) * val;
+
+                let fullscreen_from = resize
+                    .fullscreen_progress
+                    .map(|anim| anim.clamped_value().clamp(0., 1.))
+                    .unwrap_or(if prev_sizing_mode.is_fullscreen() {
+                        1.
+                    } else {
+                        0.
+                    });
+
+                let expanded_from = resize
+                    .expanded_progress
+                    .map(|anim| anim.clamped_value().clamp(0., 1.))
+                    .unwrap_or(if prev_sizing_mode.is_normal() { 0. } else { 1. });
+
                 // Also try to reuse the existing offscreen buffer if we have one.
-                (size, resize.offscreen)
+                (
+                    size,
+                    tile_size,
+                    fullscreen_from,
+                    expanded_from,
+                    resize.offscreen,
+                )
             } else {
-                (animate_from.size, OffscreenBuffer::default())
+                let size = animate_from.size;
+
+                // Compute like in tile_size().
+                let mut tile_size = size;
+                if prev_sizing_mode.is_fullscreen() {
+                    tile_size.w = f64::max(tile_size.w, self.view_size.w);
+                    tile_size.h = f64::max(tile_size.h, self.view_size.h);
+                } else if prev_sizing_mode.is_normal() && !self.border.is_off() {
+                    let width = self.border.width();
+                    tile_size.w += width * 2.;
+                    tile_size.h += width * 2.;
+                }
+
+                let fullscreen_from = if prev_sizing_mode.is_fullscreen() {
+                    1.
+                } else {
+                    0.
+                };
+
+                let expanded_from = if prev_sizing_mode.is_normal() { 0. } else { 1. };
+
+                (
+                    size,
+                    tile_size,
+                    fullscreen_from,
+                    expanded_from,
+                    OffscreenBuffer::default(),
+                )
             };
+            let (size_from, tile_size_from, fullscreen_from, expanded_from, offscreen) = params;
 
             let change = self.window.size().to_f64().to_point() - size_from.to_point();
             let change = f64::max(change.x.abs(), change.y.abs());
+            let tile_change = self.tile_size().to_f64().to_point() - tile_size_from.to_point();
+            let tile_change = f64::max(tile_change.x.abs(), tile_change.y.abs());
+            let change = f64::max(change, tile_change);
             if change > RESIZE_ANIMATION_THRESHOLD {
                 let anim = Animation::new(
                     self.clock.clone(),
@@ -270,26 +353,48 @@ impl<W: LayoutElement> Tile<W> {
                     0.,
                     self.options.animations.window_resize.anim,
                 );
+
+                let fullscreen_to = if self.sizing_mode.is_fullscreen() {
+                    1.
+                } else {
+                    0.
+                };
+                let expanded_to = if self.sizing_mode.is_normal() { 0. } else { 1. };
+                let fullscreen_progress = (fullscreen_from != fullscreen_to)
+                    .then(|| anim.restarted(fullscreen_from, fullscreen_to, 0.));
+                let expanded_progress = (expanded_from != expanded_to)
+                    .then(|| anim.restarted(expanded_from, expanded_to, 0.));
+
                 self.resize_animation = Some(ResizeAnimation {
                     anim,
                     size_from,
                     snapshot: animate_from,
                     offscreen,
+                    tile_size_from,
+                    fullscreen_progress,
+                    expanded_progress,
                 });
             } else {
                 self.resize_animation = None;
             }
         }
 
-        let rules = self.window.rules();
-        let border_config = rules.border.resolve_against(self.options.border);
-        self.border.update_config(border_config.into());
-        let focus_ring_config = rules
-            .focus_ring
-            .resolve_against(self.options.focus_ring.into());
-        self.focus_ring.update_config(focus_ring_config.into());
+        let round_max1 = |logical| round_logical_in_physical_max1(self.scale, logical);
 
-        let shadow_config = rules.shadow.resolve_against(self.options.shadow);
+        let rules = self.window.rules();
+        let mut border_config = self.options.layout.border.merged_with(&rules.border);
+        border_config.width = round_max1(border_config.width);
+        self.border.update_config(border_config.into());
+
+        let mut focus_ring_config = self
+            .options
+            .layout
+            .focus_ring
+            .merged_with(&rules.focus_ring);
+        focus_ring_config.width = round_max1(focus_ring_config.width);
+        self.focus_ring.update_config(focus_ring_config);
+
+        let shadow_config = self.options.layout.shadow.merged_with(&rules.shadow);
         self.shadow.update_config(shadow_config);
 
         let window_size = self.window_size();
@@ -349,22 +454,29 @@ impl<W: LayoutElement> Tile<W> {
 
     pub fn update_render_elements(&mut self, is_active: bool, view_rect: Rectangle<f64, Logical>) {
         let rules = self.window.rules();
+        let animated_tile_size = self.animated_tile_size();
+        let expanded_progress = self.expanded_progress();
 
         let draw_border_with_background = rules
             .draw_border_with_background
             .unwrap_or_else(|| !self.window.has_ssd());
-        let border_width = self.effective_border_width().unwrap_or(0.);
-        let radius = if self.is_fullscreen {
-            CornerRadius::default()
-        } else {
-            rules
-                .geometry_corner_radius
-                .map_or(CornerRadius::default(), |radius| {
-                    radius.expanded_by(border_width as f32)
-                })
-        };
+        let border_width = self.visual_border_width().unwrap_or(0.);
+
+        // Do the inverse of tile_size() in order to handle the unfullscreen animation for windows
+        // that were smaller than the fullscreen size, and therefore their animated_window_size() is
+        // currently much smaller than the tile size.
+        let mut border_window_size = animated_tile_size;
+        border_window_size.w -= border_width * 2.;
+        border_window_size.h -= border_width * 2.;
+
+        let radius = rules
+            .geometry_corner_radius
+            .map_or(CornerRadius::default(), |radius| {
+                radius.expanded_by(border_width as f32)
+            })
+            .scaled_by(1. - expanded_progress as f32);
         self.border.update_render_elements(
-            self.animated_window_size(),
+            border_window_size,
             is_active,
             !draw_border_with_background,
             self.window.is_urgent(),
@@ -374,40 +486,43 @@ impl<W: LayoutElement> Tile<W> {
             ),
             radius,
             self.scale,
-            1.,
+            1. - expanded_progress as f32,
         );
 
-        let radius = if self.is_fullscreen {
-            CornerRadius::default()
-        } else if self.effective_border_width().is_some() {
+        let radius = if self.visual_border_width().is_some() {
             radius
         } else {
-            rules.geometry_corner_radius.unwrap_or_default()
+            rules
+                .geometry_corner_radius
+                .unwrap_or_default()
+                .scaled_by(1. - expanded_progress as f32)
         };
         self.shadow.update_render_elements(
-            self.animated_tile_size(),
+            animated_tile_size,
             is_active,
             radius,
             self.scale,
-            1.,
+            1. - expanded_progress as f32,
         );
 
-        let draw_focus_ring_with_background = if self.effective_border_width().is_some() {
-            false
-        } else {
+        let draw_focus_ring_with_background = if self.border.is_off() {
             draw_border_with_background
+        } else {
+            false
         };
         let radius = radius.expanded_by(self.focus_ring.width() as f32);
         self.focus_ring.update_render_elements(
-            self.animated_tile_size(),
+            animated_tile_size,
             is_active,
             !draw_focus_ring_with_background,
             self.window.is_urgent(),
             view_rect,
             radius,
             self.scale,
-            1.,
+            1. - expanded_progress as f32,
         );
+
+        self.fullscreen_backdrop.resize(animated_tile_size);
     }
 
     pub fn scale(&self) -> f64 {
@@ -543,13 +658,41 @@ impl<W: LayoutElement> Tile<W> {
         &mut self.window
     }
 
-    pub fn is_fullscreen(&self) -> bool {
-        self.is_fullscreen
+    pub fn sizing_mode(&self) -> SizingMode {
+        self.sizing_mode
+    }
+
+    fn fullscreen_progress(&self) -> f64 {
+        if let Some(resize) = &self.resize_animation {
+            if let Some(anim) = &resize.fullscreen_progress {
+                return anim.clamped_value().clamp(0., 1.);
+            }
+        }
+
+        if self.sizing_mode.is_fullscreen() {
+            1.
+        } else {
+            0.
+        }
+    }
+
+    fn expanded_progress(&self) -> f64 {
+        if let Some(resize) = &self.resize_animation {
+            if let Some(anim) = &resize.expanded_progress {
+                return anim.clamped_value().clamp(0., 1.);
+            }
+        }
+
+        if self.sizing_mode.is_normal() {
+            0.
+        } else {
+            1.
+        }
     }
 
     /// Returns `None` if the border is hidden and `Some(width)` if it should be shown.
     pub fn effective_border_width(&self) -> Option<f64> {
-        if self.is_fullscreen {
+        if !self.sizing_mode.is_normal() {
             return None;
         }
 
@@ -560,33 +703,47 @@ impl<W: LayoutElement> Tile<W> {
         Some(self.border.width())
     }
 
+    fn visual_border_width(&self) -> Option<f64> {
+        if self.border.is_off() {
+            return None;
+        }
+
+        let expanded_progress = self.expanded_progress();
+
+        // Only hide the border when fully expanded to avoid jarring border appearance.
+        if expanded_progress == 1. {
+            return None;
+        }
+
+        // FIXME: would be cool to, like, gradually resize the border from full width to 0 during
+        // fullscreening, but the rest of the code isn't quite ready for that yet. It needs to
+        // handle things like computing intermediate tile size when an animated resize starts during
+        // an animated unfullscreen resize.
+        Some(self.border.width())
+    }
+
     /// Returns the location of the window's visual geometry within this Tile.
     pub fn window_loc(&self) -> Point<f64, Logical> {
         let mut loc = Point::from((0., 0.));
 
-        // In fullscreen, center the window in the given size.
-        if self.is_fullscreen {
-            let window_size = self.window_size();
-            let target_size = self.view_size;
+        let window_size = self.animated_window_size();
+        let target_size = self.animated_tile_size();
 
-            // Windows aren't supposed to be larger than the fullscreen size, but in case we get
-            // one, leave it at the top-left as usual.
-            if window_size.w < target_size.w {
-                loc.x += (target_size.w - window_size.w) / 2.;
-            }
-            if window_size.h < target_size.h {
-                loc.y += (target_size.h - window_size.h) / 2.;
-            }
+        // Center the window within its tile.
+        //
+        // - Without borders, the sizes match, so this difference is zero.
+        // - Borders always match from all sides, so this difference is pre-rounded to physical.
+        // - In fullscreen, if the window is smaller than the tile, then it gets centered, otherwise
+        //   the tile size matches the window.
+        // - During animations, the window remains centered within the tile; this is important for
+        //   the to/from fullscreen animation.
+        loc.x += (target_size.w - window_size.w) / 2.;
+        loc.y += (target_size.h - window_size.h) / 2.;
 
-            // Round to physical pixels.
-            loc = loc
-                .to_physical_precise_round(self.scale)
-                .to_logical(self.scale);
-        }
-
-        if let Some(width) = self.effective_border_width() {
-            loc += (width, width).into();
-        }
+        // Round to physical pixels.
+        loc = loc
+            .to_physical_precise_round(self.scale)
+            .to_logical(self.scale);
 
         loc
     }
@@ -594,7 +751,7 @@ impl<W: LayoutElement> Tile<W> {
     pub fn tile_size(&self) -> Size<f64, Logical> {
         let mut size = self.window_size();
 
-        if self.is_fullscreen {
+        if self.sizing_mode.is_fullscreen() {
             // Normally we'd just return the fullscreen size here, but this makes things a bit
             // nicer if a fullscreen window is bigger than the fullscreen size for some reason.
             size.w = f64::max(size.w, self.view_size.w);
@@ -613,7 +770,7 @@ impl<W: LayoutElement> Tile<W> {
     pub fn tile_expected_or_current_size(&self) -> Size<f64, Logical> {
         let mut size = self.window_expected_or_current_size();
 
-        if self.is_fullscreen {
+        if self.sizing_mode.is_fullscreen() {
             // Normally we'd just return the fullscreen size here, but this makes things a bit
             // nicer if a fullscreen window is bigger than the fullscreen size for some reason.
             size.w = f64::max(size.w, self.view_size.w);
@@ -664,19 +821,17 @@ impl<W: LayoutElement> Tile<W> {
     }
 
     pub fn animated_tile_size(&self) -> Size<f64, Logical> {
-        let mut size = self.animated_window_size();
+        let mut size = self.tile_size();
 
-        if self.is_fullscreen {
-            // Normally we'd just return the fullscreen size here, but this makes things a bit
-            // nicer if a fullscreen window is bigger than the fullscreen size for some reason.
-            size.w = f64::max(size.w, self.view_size.w);
-            size.h = f64::max(size.h, self.view_size.h);
-            return size;
-        }
+        if let Some(resize) = &self.resize_animation {
+            let val = resize.anim.value();
+            let size_from = resize.tile_size_from.to_f64();
 
-        if let Some(width) = self.effective_border_width() {
-            size.w += width * 2.;
-            size.h += width * 2.;
+            size.w = f64::max(1., size_from.w + (size.w - size_from.w) * val);
+            size.h = f64::max(1., size_from.h + (size.h - size_from.h) * val);
+            size = size
+                .to_physical_precise_round(self.scale)
+                .to_logical(self.scale);
         }
 
         size
@@ -744,8 +899,12 @@ impl<W: LayoutElement> Tile<W> {
         // The size request has to be i32 unfortunately, due to Wayland. We floor here instead of
         // round to avoid situations where proportionally-sized columns don't fit on the screen
         // exactly.
-        self.window
-            .request_size(size.to_i32_floor(), false, animate, transaction);
+        self.window.request_size(
+            size.to_i32_floor(),
+            SizingMode::Normal,
+            animate,
+            transaction,
+        );
     }
 
     pub fn tile_width_for_window_width(&self, size: f64) -> f64 {
@@ -780,9 +939,27 @@ impl<W: LayoutElement> Tile<W> {
         }
     }
 
+    pub fn request_maximized(
+        &mut self,
+        size: Size<f64, Logical>,
+        animate: bool,
+        transaction: Option<Transaction>,
+    ) {
+        self.window.request_size(
+            size.to_i32_round(),
+            SizingMode::Maximized,
+            animate,
+            transaction,
+        );
+    }
+
     pub fn request_fullscreen(&mut self, animate: bool, transaction: Option<Transaction>) {
-        self.window
-            .request_size(self.view_size.to_i32_round(), true, animate, transaction);
+        self.window.request_size(
+            self.view_size.to_i32_round(),
+            SizingMode::Fullscreen,
+            animate,
+            transaction,
+        );
     }
 
     pub fn min_size_nonfullscreen(&self) -> Size<f64, Logical> {
@@ -830,17 +1007,6 @@ impl<W: LayoutElement> Tile<W> {
         Point::from((0., y))
     }
 
-    pub fn draw_border_with_background(&self) -> bool {
-        if self.effective_border_width().is_some() {
-            return false;
-        }
-
-        self.window
-            .rules()
-            .draw_border_with_background
-            .unwrap_or_else(|| !self.window.has_ssd())
-    }
-
     fn render_inner<'a, R: NiriRenderer + 'a>(
         &'a self,
         renderer: &mut R,
@@ -851,11 +1017,17 @@ impl<W: LayoutElement> Tile<W> {
         let _span = tracy_client::span!("Tile::render_inner");
 
         let scale = Scale::from(self.scale);
+        let fullscreen_progress = self.fullscreen_progress();
+        let expanded_progress = self.expanded_progress();
 
-        let win_alpha = if self.is_fullscreen || self.window.is_ignoring_opacity_window_rule() {
+        let win_alpha = if self.window.is_ignoring_opacity_window_rule() {
             1.
         } else {
-            self.window.rules().opacity.unwrap_or(1.).clamp(0., 1.)
+            let alpha = self.window.rules().opacity.unwrap_or(1.).clamp(0., 1.);
+
+            // Interpolate towards alpha = 1. at fullscreen.
+            let p = fullscreen_progress as f32;
+            alpha * (1. - p) + 1. * p
         };
 
         // This is here rather than in render_offset() because render_offset() is currently assumed
@@ -875,8 +1047,14 @@ impl<W: LayoutElement> Tile<W> {
         let area = Rectangle::new(window_render_loc, animated_window_size);
 
         let rules = self.window.rules();
-        let clip_to_geometry = !self.is_fullscreen && rules.clip_to_geometry == Some(true);
-        let radius = rules.geometry_corner_radius.unwrap_or_default();
+
+        // Clip to geometry including during the fullscreen animation to help with buggy clients
+        // that submit a full-sized buffer before acking the fullscreen state (Firefox).
+        let clip_to_geometry = fullscreen_progress < 1. && rules.clip_to_geometry == Some(true);
+        let radius = rules
+            .geometry_corner_radius
+            .unwrap_or_default()
+            .scaled_by(1. - expanded_progress as f32);
 
         // If we're resizing, try to render a shader, or a fallback.
         let mut resize_shader = None;
@@ -970,6 +1148,7 @@ impl<W: LayoutElement> Tile<W> {
         let mut window_surface = None;
         let mut window_popups = None;
         let mut rounded_corner_damage = None;
+        let has_border_shader = BorderRenderElement::has_shader(renderer);
         if resize_shader.is_none() && resize_fallback.is_none() {
             let window = self
                 .window
@@ -979,7 +1158,6 @@ impl<W: LayoutElement> Tile<W> {
             let radius = radius.fit_to(window_size.w as f32, window_size.h as f32);
 
             let clip_shader = ClippedSurfaceRenderElement::shader(renderer).cloned();
-            let has_border_shader = BorderRenderElement::has_shader(renderer);
 
             if clip_to_geometry && clip_shader.is_some() {
                 let damage = self.rounded_corner_damage.element();
@@ -1049,28 +1227,67 @@ impl<W: LayoutElement> Tile<W> {
             .chain(rounded_corner_damage)
             .chain(window_surface.into_iter().flatten());
 
-        let elem = self.is_fullscreen.then(|| {
-            SolidColorRenderElement::from_buffer(
-                &self.fullscreen_backdrop,
-                location,
-                1.,
-                Kind::Unspecified,
-            )
-            .into()
+        let elem = (fullscreen_progress > 0.).then(|| {
+            let alpha = fullscreen_progress as f32;
+
+            // During the un/fullscreen animation, render a border element in order to use the
+            // animated corner radius.
+            if fullscreen_progress < 1. && has_border_shader {
+                let border_width = self.visual_border_width().unwrap_or(0.);
+                let radius = rules
+                    .geometry_corner_radius
+                    .map_or(CornerRadius::default(), |radius| {
+                        radius.expanded_by(border_width as f32)
+                    })
+                    .scaled_by(1. - expanded_progress as f32);
+
+                let size = self.fullscreen_backdrop.size();
+                let color = self.fullscreen_backdrop.color();
+                BorderRenderElement::new(
+                    size,
+                    Rectangle::from_size(size),
+                    GradientInterpolation::default(),
+                    Color::from_color32f(color),
+                    Color::from_color32f(color),
+                    0.,
+                    Rectangle::from_size(size),
+                    0.,
+                    radius,
+                    scale.x as f32,
+                    alpha,
+                )
+                .with_location(location)
+                .into()
+            } else {
+                SolidColorRenderElement::from_buffer(
+                    &self.fullscreen_backdrop,
+                    location,
+                    alpha,
+                    Kind::Unspecified,
+                )
+                .into()
+            }
         });
         let rv = rv.chain(elem);
 
-        let elem = self.effective_border_width().map(|width| {
+        let elem = self.visual_border_width().map(|width| {
             self.border
                 .render(renderer, location + Point::from((width, width)))
                 .map(Into::into)
         });
         let rv = rv.chain(elem.into_iter().flatten());
 
-        let elem = focus_ring.then(|| self.focus_ring.render(renderer, location).map(Into::into));
+        // Hide the focus ring when maximized/fullscreened. It's not normally visible anyway due to
+        // being outside the monitor or obscured by a solid colored bar, but it is visible under
+        // semitransparent bars in maximized state (which is a bit weird) and in the overview (also
+        // a bit weird).
+        let elem = (focus_ring && expanded_progress < 1.)
+            .then(|| self.focus_ring.render(renderer, location).map(Into::into));
         let rv = rv.chain(elem.into_iter().flatten());
 
-        rv.chain(self.shadow.render(renderer, location).map(Into::into))
+        let elem = (expanded_progress < 1.)
+            .then(|| self.shadow.render(renderer, location).map(Into::into));
+        rv.chain(elem.into_iter().flatten())
     }
 
     pub fn render<'a, R: NiriRenderer + 'a>(
@@ -1199,8 +1416,7 @@ impl<W: LayoutElement> Tile<W> {
     pub fn verify_invariants(&self) {
         use approx::assert_abs_diff_eq;
 
-        assert_eq!(self.is_fullscreen, self.window.is_fullscreen());
-        assert_eq!(self.fullscreen_backdrop.size(), self.view_size);
+        assert_eq!(self.sizing_mode, self.window.sizing_mode());
 
         let scale = self.scale;
         let size = self.tile_size();
