@@ -1,6 +1,8 @@
 #[macro_use]
 extern crate tracing;
 
+#[cfg(feature = "dbus")]
+pub mod a11y;
 pub mod animation;
 pub mod backend;
 pub mod cli;
@@ -11,6 +13,7 @@ pub mod frame_clock;
 pub mod handlers;
 pub mod input;
 pub mod ipc;
+pub mod layer;
 pub mod layout;
 pub mod niri;
 pub mod protocols;
@@ -27,3 +30,6 @@ pub mod pw_utils;
 
 #[cfg(not(feature = "xdp-gnome-screencast"))]
 pub use dummy_pw_utils as pw_utils;
+
+#[cfg(test)]
+mod tests;
