@@ -519,7 +519,7 @@ where
                     }
                 };
                 pointer.mutate_axis_frame(Some(time), |frame| {
-                    frame.value(axis, value).v120(axis, discrete)
+                    frame.value(axis, value).v120(axis, discrete * 120)
                 });
             }
             zwlr_virtual_pointer_v1::Request::Destroy => {}
