@@ -14,6 +14,7 @@ pub struct Shaders {
     pub shadow: Option<ShaderProgram>,
     pub clipped_surface: Option<GlesTexProgram>,
     pub resize: Option<ShaderProgram>,
+    pub gradient_fade: Option<GlesTexProgram>,
     pub custom_resize: RefCell<Option<ShaderProgram>>,
     pub custom_close: RefCell<Option<ShaderProgram>>,
     pub custom_open: RefCell<Option<ShaderProgram>>,
@@ -96,11 +97,22 @@ impl Shaders {
             })
             .ok();
 
+        let gradient_fade = renderer
+            .compile_custom_texture_shader(
+                include_str!("gradient_fade.frag"),
+                &[UniformName::new("cutoff", UniformType::_2f)],
+            )
+            .map_err(|err| {
+                warn!("error compiling gradient fade shader: {err:?}");
+            })
+            .ok();
+
         Self {
             border,
             shadow,
             clipped_surface,
             resize,
+            gradient_fade,
             custom_resize: RefCell::new(None),
             custom_close: RefCell::new(None),
             custom_open: RefCell::new(None),
