@@ -25,7 +25,8 @@ use smithay::utils::{Coordinate, Logical, Point, Rectangle, Size, Transform};
 use smithay::wayland::compositor::{send_surface_state, with_states, SurfaceData};
 use smithay::wayland::fractional_scale::with_fractional_scale;
 use smithay::wayland::shell::xdg::{
-    ToplevelSurface, XdgToplevelSurfaceData, XdgToplevelSurfaceRoleAttributes,
+    ToplevelCachedState, ToplevelConfigure, ToplevelState, ToplevelSurface, XdgToplevelSurfaceData,
+    XdgToplevelSurfaceRoleAttributes,
 };
 use wayland_backend::server::Credentials;
 
@@ -37,6 +38,7 @@ pub mod scale;
 pub mod signals;
 pub mod spawning;
 pub mod transaction;
+pub mod vblank_throttle;
 pub mod watcher;
 pub mod xwayland;
 
@@ -172,6 +174,15 @@ pub fn logical_output(output: &Output) -> niri_ipc::LogicalOutput {
     }
 }
 
+pub struct PanelOrientation(pub Transform);
+pub fn panel_orientation(output: &Output) -> Transform {
+    output
+        .user_data()
+        .get::<PanelOrientation>()
+        .map(|x| x.0)
+        .unwrap_or(Transform::Normal)
+}
+
 pub fn ipc_transform_to_smithay(transform: niri_ipc::Transform) -> Transform {
     match transform {
         niri_ipc::Transform::Normal => Transform::Normal,
@@ -212,7 +223,7 @@ pub fn expand_home(path: &Path) -> anyhow::Result<Option<PathBuf>> {
 }
 
 pub fn make_screenshot_path(config: &Config) -> anyhow::Result<Option<PathBuf>> {
-    let Some(path) = &config.screenshot_path else {
+    let Some(path) = &config.screenshot_path.0 else {
         return Ok(None);
     };
 
@@ -279,6 +290,60 @@ pub fn with_toplevel_role<T>(
     })
 }
 
+pub fn with_toplevel_role_and_current<T>(
+    toplevel: &ToplevelSurface,
+    f: impl FnOnce(&mut XdgToplevelSurfaceRoleAttributes, Option<&ToplevelState>) -> T,
+) -> T {
+    with_states(toplevel.wl_surface(), |states| {
+        let mut role = states
+            .data_map
+            .get::<XdgToplevelSurfaceData>()
+            .unwrap()
+            .lock()
+            .unwrap();
+
+        let mut guard = states.cached_state.get::<ToplevelCachedState>();
+        let current = guard.current().last_acked.as_ref().map(|c| &c.state);
+
+        f(&mut role, current)
+    })
+}
+
+pub fn with_toplevel_last_uncommitted_configure<T>(
+    toplevel: &ToplevelSurface,
+    f: impl FnOnce(Option<&ToplevelConfigure>) -> T,
+) -> T {
+    with_states(toplevel.wl_surface(), |states| {
+        let role = states
+            .data_map
+            .get::<XdgToplevelSurfaceData>()
+            .unwrap()
+            .lock()
+            .unwrap();
+
+        let mut guard = states.cached_state.get::<ToplevelCachedState>();
+
+        if let Some(last_pending) = role.pending_configures().last() {
+            // Configure not yet acked by the client.
+            f(Some(last_pending))
+        } else if let Some(last_acked) = &role.last_acked {
+            let mut configure = Some(last_acked);
+
+            if let Some(committed) = &guard.current().last_acked {
+                if committed.serial.is_no_older_than(&last_acked.serial) {
+                    // Already committed to this configure.
+                    configure = None;
+                }
+            }
+
+            f(configure)
+        } else {
+            // Surface hadn't been configured yet.
+            f(None)
+        }
+    })
+}
+
 pub fn update_tiled_state(
     toplevel: &ToplevelSurface,
     prefer_no_csd: bool,
@@ -406,7 +471,7 @@ pub fn baba_is_float_offset(now: Duration, view_height: f64) -> f64 {
 }
 
 #[cfg(feature = "dbus")]
-pub fn show_screenshot_notification(image_path: Option<PathBuf>) -> anyhow::Result<()> {
+pub fn show_screenshot_notification(image_path: Option<&Path>) -> anyhow::Result<()> {
     use std::collections::HashMap;
 
     use zbus::zvariant;
