@@ -15,14 +15,17 @@ use smithay::reexports::wayland_server::Resource as _;
 use smithay::utils::{Logical, Point, Rectangle, Scale, Serial, Size, Transform};
 use smithay::wayland::compositor::{remove_pre_commit_hook, with_states, HookId, SurfaceData};
 use smithay::wayland::seat::WaylandFocus;
-use smithay::wayland::shell::xdg::{SurfaceCachedState, ToplevelSurface};
+use smithay::wayland::shell::xdg::{
+    SurfaceCachedState, ToplevelCachedState, ToplevelConfigure, ToplevelSurface,
+    XdgToplevelSurfaceData,
+};
 use wayland_backend::server::Credentials;
 
 use super::{ResolvedWindowRules, WindowRef};
 use crate::handlers::KdeDecorationsModeState;
 use crate::layout::{
     ConfigureIntent, InteractiveResizeData, LayoutElement, LayoutElementRenderElement,
-    LayoutElementRenderSnapshot,
+    LayoutElementRenderSnapshot, SizingMode,
 };
 use crate::niri_render_elements;
 use crate::render_helpers::border::BorderRenderElement;
@@ -35,7 +38,8 @@ use crate::render_helpers::{BakedBuffer, RenderTarget, SplitElements};
 use crate::utils::id::IdCounter;
 use crate::utils::transaction::Transaction;
 use crate::utils::{
-    get_credentials_for_surface, send_scale_transform, update_tiled_state, with_toplevel_role,
+    get_credentials_for_surface, send_scale_transform, update_tiled_state,
+    with_toplevel_last_uncommitted_configure, with_toplevel_role, with_toplevel_role_and_current,
     ResizeEdge,
 };
 
@@ -160,6 +164,28 @@ pub struct Mapped {
     /// These have been "sent" to the window in form of configures, but the window hadn't committed
     /// in response yet.
     uncommitted_windowed_fullscreen: Vec<(Serial, bool)>,
+
+    /// Whether this window is maximized.
+    ///
+    /// We have to track this ourselves in addition to the Maximized toplevel state in order to
+    /// support windowed fullscreen, since in windowed fullscreen the toplevel state is always
+    /// Fullscreen. So we need this variable to be able to report accurate sizing mode and pending
+    /// sizing mode.
+    is_maximized: bool,
+
+    /// Whether this window is pending to be maximized.
+    ///
+    /// We have to track this ourselves due to windowed fullscreen.
+    is_pending_maximized: bool,
+
+    /// Pending maximized updates.
+    ///
+    /// These have been "sent" to the window in form of configures, but the window hadn't committed
+    /// in response yet.
+    uncommitted_maximized: Vec<(Serial, bool)>,
+
+    /// Most recent monotonic time when the window had the focus.
+    focus_timestamp: Option<Duration>,
 }
 
 niri_render_elements! {
@@ -225,7 +251,7 @@ impl Mapped {
         let surface = window.wl_surface().expect("no X11 support");
         let credentials = get_credentials_for_surface(&surface);
 
-        Self {
+        let mut rv = Self {
             window,
             id: MappedId::next(),
             credentials,
@@ -253,7 +279,16 @@ impl Mapped {
             is_windowed_fullscreen: false,
             is_pending_windowed_fullscreen: false,
             uncommitted_windowed_fullscreen: Vec::new(),
-        }
+            is_maximized: false,
+            is_pending_maximized: false,
+            uncommitted_maximized: Vec::new(),
+            focus_timestamp: None,
+        };
+
+        rv.is_maximized = rv.sizing_mode().is_maximized();
+        rv.is_pending_maximized = rv.pending_sizing_mode().is_maximized();
+
+        rv
     }
 
     pub fn toplevel(&self) -> &ToplevelSurface {
@@ -484,6 +519,14 @@ impl Mapped {
         })
     }
 
+    pub fn get_focus_timestamp(&self) -> Option<Duration> {
+        self.focus_timestamp
+    }
+
+    pub fn set_focus_timestamp(&mut self, timestamp: Duration) {
+        self.focus_timestamp.replace(timestamp);
+    }
+
     pub fn send_frame<T, F>(
         &mut self,
         output: &Output,
@@ -667,12 +710,12 @@ impl LayoutElement for Mapped {
     fn request_size(
         &mut self,
         size: Size<i32, Logical>,
-        is_fullscreen: bool,
+        mode: SizingMode,
         animate: bool,
         transaction: Option<Transaction>,
     ) {
         // Going into real fullscreen resets windowed fullscreen.
-        if is_fullscreen {
+        if mode == SizingMode::Fullscreen {
             self.is_pending_windowed_fullscreen = false;
 
             if self.is_windowed_fullscreen {
@@ -682,14 +725,27 @@ impl LayoutElement for Mapped {
             }
         }
 
+        self.is_pending_maximized = mode == SizingMode::Maximized;
+        if self.is_maximized != self.is_pending_maximized {
+            // Make sure we receive a commit to update self.is_maximized later on.
+            self.needs_configure = true;
+        }
+
         let changed = self.toplevel().with_pending_state(|state| {
             let changed = state.size != Some(size);
             state.size = Some(size);
-            if is_fullscreen || self.is_pending_windowed_fullscreen {
+
+            if mode.is_fullscreen() || self.is_pending_windowed_fullscreen {
                 state.states.set(xdg_toplevel::State::Fullscreen);
+                state.states.unset(xdg_toplevel::State::Maximized);
+            } else if mode.is_maximized() {
+                state.states.unset(xdg_toplevel::State::Fullscreen);
+                state.states.set(xdg_toplevel::State::Maximized);
             } else {
                 state.states.unset(xdg_toplevel::State::Fullscreen);
+                state.states.unset(xdg_toplevel::State::Maximized);
             }
+
             changed
         });
 
@@ -714,34 +770,39 @@ impl LayoutElement for Mapped {
         // longer participate in any transactions with other windows.
         self.transaction_for_next_configure = None;
 
+        self.is_pending_maximized = false;
+        if self.is_maximized != self.is_pending_maximized {
+            // Make sure we receive a commit to update self.is_maximized later on.
+            self.needs_configure = true;
+        }
+
         // If our last requested size already matches the size we want to request-once, clear the
         // size request right away. However, we must also check if we're unfullscreening, because
         // in that case the window itself will restore its previous size upon receiving a (0, 0)
         // configure, whereas what we potentially want is to unfullscreen the window into its
         // fullscreen size.
-        let already_sent = with_toplevel_role(self.toplevel(), |role| {
-            let (last_sent, last_serial) = if let Some(configure) = role.pending_configures().last()
-            {
-                // FIXME: it would be more optimal to find the *oldest* pending configure that
-                // has the same size and fullscreen state to the last pending configure.
-                (&configure.state, configure.serial)
-            } else {
-                (
-                    role.last_acked.as_ref().unwrap(),
-                    role.configure_serial.unwrap(),
-                )
-            };
+        let already_sent = with_toplevel_last_uncommitted_configure(self.toplevel(), |configure| {
+            let ToplevelConfigure { state, serial } = configure?;
 
-            let same_size = last_sent.size.unwrap_or_default() == size;
-            let has_fullscreen = last_sent.states.contains(xdg_toplevel::State::Fullscreen);
+            let same_size = state.size.unwrap_or_default() == size;
+            let has_fullscreen = state.states.contains(xdg_toplevel::State::Fullscreen);
             let same_fullscreen = has_fullscreen == self.is_pending_windowed_fullscreen;
-            (same_size && same_fullscreen).then_some(last_serial)
+            let has_maximized = state.states.contains(xdg_toplevel::State::Maximized);
+            let same_maximized = !has_maximized;
+            (same_size && same_fullscreen && same_maximized).then_some(*serial)
         });
 
         if let Some(serial) = already_sent {
-            if let Some(current_serial) =
-                with_toplevel_role(self.toplevel(), |role| role.current_serial)
-            {
+            let current_serial = with_states(self.toplevel().wl_surface(), |states| {
+                states
+                    .cached_state
+                    .get::<ToplevelCachedState>()
+                    .current()
+                    .last_acked
+                    .as_ref()
+                    .map(|c| c.serial)
+            });
+            if let Some(current_serial) = current_serial {
                 // God this triple negative...
                 if !current_serial.is_no_older_than(&serial) {
                     // We have already sent a request for the new size, but the surface has not
@@ -765,6 +826,7 @@ impl LayoutElement for Mapped {
             if !self.is_pending_windowed_fullscreen {
                 state.states.unset(xdg_toplevel::State::Fullscreen);
             }
+            state.states.unset(xdg_toplevel::State::Maximized);
             changed
         });
 
@@ -805,7 +867,9 @@ impl LayoutElement for Mapped {
 
     fn has_ssd(&self) -> bool {
         let toplevel = self.toplevel();
-        let mode = with_toplevel_role(self.toplevel(), |role| role.current.decoration_mode);
+        let mode = self
+            .toplevel()
+            .with_committed_state(|current| current.and_then(|s| s.decoration_mode));
 
         match mode {
             Some(zxdg_toplevel_decoration_v1::Mode::ServerSide) => true,
@@ -892,10 +956,10 @@ impl LayoutElement for Mapped {
             return ConfigureIntent::ShouldSend;
         }
 
-        with_toplevel_role(self.toplevel(), |attributes| {
+        with_toplevel_role_and_current(self.toplevel(), |attributes, current_committed| {
             if let Some(server_pending) = &attributes.server_pending {
                 let current_server = attributes.current_server_state();
-                if server_pending != current_server {
+                if *server_pending != current_server {
                     // Something changed. Check if the only difference is the size, and if the
                     // current server size matches the current committed size.
                     let mut current_server_same_size = current_server.clone();
@@ -903,12 +967,17 @@ impl LayoutElement for Mapped {
                     if current_server_same_size == *server_pending {
                         // Only the size changed. Check if the window committed our previous size
                         // request.
-                        if attributes.current.size == current_server.size {
+                        let Some(current_committed) = current_committed else {
+                            error!("mapped must have had initial commit");
+                            return ConfigureIntent::ShouldSend;
+                        };
+
+                        if current_committed.size == current_server.size {
                             // The window had committed for our previous size change, so we can
                             // change the size again.
                             trace!(
                                 "current size matches server size: {:?}",
-                                attributes.current.size
+                                current_committed.size
                             );
                             ConfigureIntent::CanSend
                         } else {
@@ -960,7 +1029,7 @@ impl LayoutElement for Mapped {
                 }
 
                 let server_pending = role.server_pending.as_ref().unwrap();
-                server_pending != role.current_server_state()
+                *server_pending != role.current_server_state()
             });
 
         if has_pending_changes {
@@ -1013,6 +1082,18 @@ impl LayoutElement for Mapped {
                 self.uncommitted_windowed_fullscreen
                     .push((serial, self.is_pending_windowed_fullscreen));
             }
+
+            // If is_pending_maximized changed compared to the last value that we "sent" to the
+            // window, store the configure serial.
+            let last_sent_maximized = self
+                .uncommitted_maximized
+                .last()
+                .map(|(_, value)| *value)
+                .unwrap_or(self.is_maximized);
+            if last_sent_maximized != self.is_pending_maximized {
+                self.uncommitted_maximized
+                    .push((serial, self.is_pending_maximized));
+            }
         } else {
             self.interactive_resize = match self.interactive_resize.take() {
                 // We probably started and stopped resizing in the same loop cycle without anything
@@ -1026,25 +1107,51 @@ impl LayoutElement for Mapped {
         self.transaction_for_next_configure = None;
     }
 
-    fn is_fullscreen(&self) -> bool {
+    fn sizing_mode(&self) -> SizingMode {
         if self.is_windowed_fullscreen {
-            return false;
+            return if self.is_maximized {
+                SizingMode::Maximized
+            } else {
+                SizingMode::Normal
+            };
         }
 
-        with_toplevel_role(self.toplevel(), |role| {
-            role.current
-                .states
-                .contains(xdg_toplevel::State::Fullscreen)
+        self.toplevel().with_committed_state(|state| {
+            // This must always be Some() for mapped windows. However, this function is called on
+            // the code path when removing a just-unmapped window in the commit handler, at which
+            // point state is already None.
+            let Some(state) = state else {
+                return SizingMode::Normal;
+            };
+
+            if state.states.contains(xdg_toplevel::State::Fullscreen) {
+                SizingMode::Fullscreen
+            } else if state.states.contains(xdg_toplevel::State::Maximized) {
+                SizingMode::Maximized
+            } else {
+                SizingMode::Normal
+            }
         })
     }
 
-    fn is_pending_fullscreen(&self) -> bool {
+    fn pending_sizing_mode(&self) -> SizingMode {
         if self.is_pending_windowed_fullscreen {
-            return false;
+            return if self.is_pending_maximized {
+                SizingMode::Maximized
+            } else {
+                SizingMode::Normal
+            };
         }
 
-        self.toplevel()
-            .with_pending_state(|state| state.states.contains(xdg_toplevel::State::Fullscreen))
+        self.toplevel().with_pending_state(|state| {
+            if state.states.contains(xdg_toplevel::State::Fullscreen) {
+                SizingMode::Fullscreen
+            } else if state.states.contains(xdg_toplevel::State::Maximized) {
+                SizingMode::Maximized
+            } else {
+                SizingMode::Normal
+            }
+        })
     }
 
     fn is_ignoring_opacity_window_rule(&self) -> bool {
@@ -1056,8 +1163,8 @@ impl LayoutElement for Mapped {
     }
 
     fn expected_size(&self) -> Option<Size<i32, Logical>> {
-        // We can only use current size if it's not fullscreen.
-        let current_size = (!self.is_fullscreen()).then(|| self.window.geometry().size);
+        // We can only use current size if it's not maximized or fullscreen.
+        let current_size = (self.sizing_mode().is_normal()).then(|| self.window.geometry().size);
 
         // Check if we should be using the current window size.
         //
@@ -1076,46 +1183,54 @@ impl LayoutElement for Mapped {
             return current_size;
         }
 
-        let pending = with_toplevel_role(self.toplevel(), |role| {
+        let pending = with_states(self.toplevel().wl_surface(), |states| {
+            let role = states
+                .data_map
+                .get::<XdgToplevelSurfaceData>()
+                .unwrap()
+                .lock()
+                .unwrap();
+
             // If we have a server-pending size change that we haven't sent yet, use that size.
-            if let Some(server_pending) = &role.server_pending {
-                let current_server = role.current_server_state();
-                if server_pending.size != current_server.size {
-                    return Some((
-                        server_pending.size.unwrap_or_default(),
-                        server_pending
-                            .states
-                            .contains(xdg_toplevel::State::Fullscreen),
-                    ));
-                }
-            }
+            let server_pending = role.server_pending.as_ref()?;
 
-            // If we have a sent-but-not-committed-to size, use that.
-            let (last_sent, last_serial) = if let Some(configure) = role.pending_configures().last()
-            {
-                (&configure.state, configure.serial)
-            } else {
-                (
-                    role.last_acked.as_ref().unwrap(),
-                    role.configure_serial.unwrap(),
-                )
-            };
-
-            if let Some(current_serial) = role.current_serial {
-                if !current_serial.is_no_older_than(&last_serial) {
-                    return Some((
-                        last_sent.size.unwrap_or_default(),
-                        last_sent.states.contains(xdg_toplevel::State::Fullscreen),
-                    ));
-                }
+            let current_server = role.current_server_state();
+            if server_pending.size != current_server.size {
+                return Some((
+                    server_pending.size.unwrap_or_default(),
+                    server_pending
+                        .states
+                        .contains(xdg_toplevel::State::Fullscreen),
+                    server_pending
+                        .states
+                        .contains(xdg_toplevel::State::Maximized),
+                ));
             }
 
             None
+        })
+        .or_else(|| {
+            with_toplevel_last_uncommitted_configure(self.toplevel(), |configure| {
+                // If we have a sent-but-not-committed-to size, use that.
+                let ToplevelConfigure { state, .. } = configure?;
+
+                Some((
+                    state.size.unwrap_or_default(),
+                    state.states.contains(xdg_toplevel::State::Fullscreen),
+                    state.states.contains(xdg_toplevel::State::Maximized),
+                ))
+            })
         });
 
-        if let Some((mut size, fullscreen)) = pending {
-            // If the pending change is fullscreen, we can't use that size.
-            if fullscreen && !self.is_pending_windowed_fullscreen {
+        if let Some((mut size, fullscreen, maximized)) = pending {
+            // If the pending change is maximized or fullscreen, we can't use that size.
+            //
+            // Pending windowed fullscreen is good (means not real fullscreen), unless it's also
+            // pending maximized (means maximized windowed fullscreen, so maximized size, bad).
+            if maximized
+                || (fullscreen
+                    && (!self.is_pending_windowed_fullscreen || self.is_pending_maximized))
+            {
                 return None;
             }
 
@@ -1153,8 +1268,13 @@ impl LayoutElement for Mapped {
         self.toplevel().with_pending_state(|state| {
             if value {
                 state.states.set(xdg_toplevel::State::Fullscreen);
+                state.states.unset(xdg_toplevel::State::Maximized);
             } else {
                 state.states.unset(xdg_toplevel::State::Fullscreen);
+
+                if self.is_pending_maximized {
+                    state.states.set(xdg_toplevel::State::Maximized);
+                }
             }
         });
 
@@ -1233,5 +1353,15 @@ impl LayoutElement for Mapped {
                     true
                 }
             });
+
+        // "Commit" our "acked" pending maximized state.
+        self.uncommitted_maximized.retain_mut(|(serial, value)| {
+            if commit_serial.is_no_older_than(serial) {
+                self.is_maximized = *value;
+                false
+            } else {
+                true
+            }
+        });
     }
 }
