@@ -1,5 +1,6 @@
 use std::cmp::{max, min};
 
+use niri_config::utils::MergeWith as _;
 use niri_config::window_rule::{Match, WindowRule};
 use niri_config::{
     BlockOutFrom, BorderRule, CornerRadius, FloatingPosition, PresetSize, ShadowRule,
@@ -29,7 +30,7 @@ pub enum WindowRef<'a> {
 }
 
 /// Rules fully resolved for a window.
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Default, PartialEq, Clone)]
 pub struct ResolvedWindowRules {
     /// Default width for this window.
     ///
@@ -60,6 +61,9 @@ pub struct ResolvedWindowRules {
     /// Whether the window should open full-width.
     pub open_maximized: Option<bool>,
 
+    /// Whether the window should open maximized to edges (true maximized).
+    pub open_maximized_to_edges: Option<bool>,
+
     /// Whether the window should open fullscreen.
     pub open_fullscreen: Option<bool>,
 
@@ -170,78 +174,10 @@ impl<'a> WindowRef<'a> {
 }
 
 impl ResolvedWindowRules {
-    pub const fn empty() -> Self {
-        Self {
-            default_width: None,
-            default_height: None,
-            default_column_display: None,
-            default_floating_position: None,
-            open_on_output: None,
-            open_on_workspace: None,
-            open_maximized: None,
-            open_fullscreen: None,
-            open_floating: None,
-            open_focused: None,
-            min_width: None,
-            min_height: None,
-            max_width: None,
-            max_height: None,
-            focus_ring: BorderRule {
-                off: false,
-                on: false,
-                width: None,
-                active_color: None,
-                inactive_color: None,
-                urgent_color: None,
-                active_gradient: None,
-                inactive_gradient: None,
-                urgent_gradient: None,
-            },
-            border: BorderRule {
-                off: false,
-                on: false,
-                width: None,
-                active_color: None,
-                inactive_color: None,
-                urgent_color: None,
-                active_gradient: None,
-                inactive_gradient: None,
-                urgent_gradient: None,
-            },
-            shadow: ShadowRule {
-                off: false,
-                on: false,
-                offset: None,
-                softness: None,
-                spread: None,
-                draw_behind_window: None,
-                color: None,
-                inactive_color: None,
-            },
-            tab_indicator: TabIndicatorRule {
-                active_color: None,
-                inactive_color: None,
-                urgent_color: None,
-                active_gradient: None,
-                inactive_gradient: None,
-                urgent_gradient: None,
-            },
-            draw_border_with_background: None,
-            opacity: None,
-            geometry_corner_radius: None,
-            clip_to_geometry: None,
-            baba_is_float: None,
-            block_out_from: None,
-            variable_refresh_rate: None,
-            scroll_factor: None,
-            tiled_state: None,
-        }
-    }
-
     pub fn compute(rules: &[WindowRule], window: WindowRef, is_at_startup: bool) -> Self {
         let _span = tracy_client::span!("ResolvedWindowRules::compute");
 
-        let mut resolved = ResolvedWindowRules::empty();
+        let mut resolved = ResolvedWindowRules::default();
 
         with_toplevel_role(window.toplevel(), |role| {
             // Ensure server_pending like in Smithay's with_pending_state().
@@ -299,6 +235,10 @@ impl ResolvedWindowRules {
                     resolved.open_maximized = Some(x);
                 }
 
+                if let Some(x) = rule.open_maximized_to_edges {
+                    resolved.open_maximized_to_edges = Some(x);
+                }
+
                 if let Some(x) = rule.open_fullscreen {
                     resolved.open_fullscreen = Some(x);
                 }
