@@ -21,6 +21,9 @@ pub enum InitialConfigureState {
     NotConfigured {
         /// Whether the window requested to be fullscreened, and the requested output, if any.
         wants_fullscreen: Option<Option<Output>>,
+
+        /// Whether the window requested to be maximized.
+        wants_maximized: bool,
     },
     /// The window has been configured.
     Configured {
@@ -64,6 +67,13 @@ pub enum InitialConfigureState {
 
         /// Workspace to open this window on.
         workspace_name: Option<String>,
+
+        /// Whether the window should be maximized.
+        ///
+        /// This corresponds to the window having the Maximized toplevel state. However, if the
+        /// window is also pending fullscreen, then it has the Fullscreen toplevel state, so we
+        /// need to store pending maximized elsewhere, hence this field.
+        is_pending_maximized: bool,
     },
 }
 
@@ -74,6 +84,7 @@ impl Unmapped {
             window,
             state: InitialConfigureState::NotConfigured {
                 wants_fullscreen: None,
+                wants_maximized: false,
             },
             activation_token_data: None,
         }
